// XASM 0.1 xCode assembler

#define writeop(x) (*otp++ = (x))

#include "xasm.h"

char n_args, acode[MAX_CLINE], osz, oos;
unsigned char out[MAX_OUT], *otp = out, outfile[BS];
labelt *label;
labelrt *labelr;
datat *data;
int n_label = 0, n_labelr = 0, n_data = 0;
char opcode[BS], (*arg)[ARGSZ];
int line = 0, errcount = 0, n_arg = START_NARG;
deft *def;
int n_defs = 0;

FILE *in;

char *_strlwr(char *str)
{
	char *sp = str;

	for (; *sp; ++sp)
		*sp = tolower(*sp);
	return str;
}

_int64 atoi64(char *str)
{
	_int64 ret = 0, sign = 1;

	if (*str == '-')
	{
		sign = -1;
		++str;
	}
	for (; *str; ++str)
	{
		ret *= 10;
		ret += *str - '0';
	}
	return ret * sign;
}

char *trim_spaces(char *text)
{
	short go = 0;
	int j, s = strlen(text), c = -1;
	char *old_text;

	old_text = malloc(strlen(text)+1);
	strcpy(old_text, text);

	text[0] = '\0';
	for (j = 0; j < s; ++j)
	{
		if (!go && old_text[j] != ' ' && old_text[j] != '\t')
		{
			text[++c] = old_text[j];
			go = 1;
		}
		else if (go)
			text[++c] = old_text[j];
	}
	text[++c] = '\0';
	s = strlen(text) - 1;
	for (j = s; j >= 0; --j)
	{
		if (text[j] != ' ' && text[j] != '\t')
		{
			text[j+1] = '\0';
			goto L1;
		}
	}

L1:	free(old_text);
	return text;
}

void exiterr(char *text)
{
	puts(text);
	exit(1);
}

char *strcpy2(char *to, char *from)
{
	if (from)
		strcpy(to, from);
	else
		*to = '\0';

	return to;
}

void print_err(char *text)
{
	printf("%5d: %s\n", line, text);
	++errcount;
}

char wrong_argnum(char *opcode, int num, int target)
{
	char err[BS];

	if (num != target)
	{
		sprintf(err, "%d args given for %s but %d are expected", num, opcode, target);
		return 1;
	}

	return 0;
}

char is_int(char *str)
{
	int j, s = strlen(str);

	if (!s) return 0;
	if (*str != '-' && !isdigit(*str)) return 0;
	if (*str == '-' && !str[1]) return 0;
	for (j = 1; j < s; ++j)
		if (!isdigit(str[j]))
			return 0;
	return 1;
}

char is_float(char *str)
{
	int j, s = strlen(str), point = (*str == '.');
	char e = 0;

	if (!s) return 0;
	if (*str != '-' && !isdigit(*str) && *str != '.') return 0;
	if ((*str == '-' || *str == '.') && !str[1]) return 0;
	for (j = 1; j < s; ++j)
		if (e)
		{
			if (!isdigit(str[j]) && !((str[j] == '+' || str[j] == '-')
				&& e == 1))
				return 0;
			e = 2;
		}
		else
		{
		if (str[j] == '.')
			++point;
		else if (str[j] == 'E' || str[j] == 'e')
			e = 1;
		else if (!isdigit(str[j]))
			return 0;
		}
	return point == 1 || (e && point <= 1);
}

char is_reg(char *str)
{
	if (*str >= 'a' && *str <= 'p' && !str[1])
		return 2;
	return 0;
}

char is_lptr(char *str)
{
	return str[0] == '(' && str[strlen(str)-1] == ')';
}

char is_label(char *str)
{
	if ((!isalpha(*str) && *str != '_') || is_reg(str)) return 0;

	for (++str; *str; ++str)
		if (!isalnum(*str) && *str != '_')
			return 0;

	return 1;
}

void squish(char *str)
{
	int j, s = strlen(str)-2;

	for (j = 0; j < s; ++j)
		str[j] = str[j+1];
	str[s] = '\0';
}

void pack_int(char *str, char len)
{
	unsigned _int64 i;
	char j;

	if (*str == '-')
		i = pow(2, 8*len) - atoi64(&str[1]);
	else
		i = atoi64(str);
	for (j = 0; j < len; ++j)
	{
		*otp++ = i & 255;
		i >>= 8;
	}
}

void print_bin(char *str, int len)
{
	int j, k;
	unsigned char ch;

	printf("Binary:");
	for (j = len-1; j >= 0; --j)
	{
		putchar(' ');
		ch = str[j];
		for (k = 0; k < 8; ++k)
		{
			printf("%d", !!(ch & 128));
			ch <<= 1;
		}
	}
	putchar('\n');
}

int log2(unsigned _int64 i)
{
#ifdef WIN32
	unsigned _int64 cb = 0x8000000000000000;
#else
	unsigned _int64 cb = 0x8000000000000000ULL;
#endif
	int j;

	for (j = 63; j >= 0; --j)
	{
		if (i & cb)
			return j;
		cb >>= 1;
	}
	return 0;
}

void pack_float(char *str, char len)
{
	unsigned char sign = 0, mode = 0, wait, *os = otp;
	unsigned int n;
	unsigned _int64 m;
	double f, fm = 0;
	unsigned int j, k, sz, mn = mantissa[len], en = exponent[len];
	int e, em = 0, cm = 0;
	int len2 = tsz[len];

	if (*str == '-')
	{
		sign = 128;
		++str;
	}
	sz = strlen(str);

	for (j = 0; j < sz; ++j)
		if (str[j] == '.')
		{
			for (k = j+1; k < sz; ++k)
				if (str[k] == 'E' || str[k] == 'e')
				{
					f = atoi(&str[k+1]) / LOG_10_2;
					em = f;
					fm = fmod(f, 1);
					cm = atoi64(str) ^ (1 << em);
					str[k] = '\0';
					break;
				}
			f = atof(&str[j]);
			str[j] = '\0';
			n = atoi64(str);
			break;
		}
		else if (str[j] == 'E' || str[j] == 'e')
		{
			f = atoi(&str[j+1]) / LOG_10_2;
			em = f;
			fm = fmod(f, 1);
			str[j] = '\0';
		}
	if (j == sz)
	{
		f = 0.0;
		n = atoi64(str);
	}

	if (!n && !f)
	{
		e = 0;
		m = 0;
	}
	else
	{

		fm = pow(2, fm);
		f = f * fm + fmod(n * fm, 1);
		n *= fm;
		if (f > 1)
		{
			--f;
			++n;
		}
		if (n)
		{
			e = log2(n);
			m = n ^ (1 << e);
			sz = mn - e - 1;
		}
		else
		{
			e = -1;
			m = 0;
			sz = mn;
		}
		e += em;

		wait = !n;
		m <<= 1;
		if (!f)
			m <<= sz;
		else for (j = 0; j < sz; ++j)
		{
			f *= 2;
			if (f >= 1)
			{
				if (!wait)
					++m;
				--f;
				wait = 0;
			}
			else if (wait)
			{
				--e;
				++sz;
			}
			m <<= 1;
		}
		if (f * 4 >= 1)
			++m;

		e += bias[len];
	}

	for (j = 0; j < len2; ++j)
	{
		switch (mode)
		{
		case 0:
			if (mn >= 8)
			{
				*otp++ = m & 255;
				m >>= 8;
				mn -= 8;
			}
			else
			{
				*otp++ = ((e & (255 >> mn)) << mn) | m;
				en -= 8 - mn;
				e >>= 8 - mn;
				mode = 1;
			}
			break;
		case 1:
			if (en >= 8)
			{
				*otp++ = e & 255;
				e >>= 8;
				en -= 8;
			}
			else
			{
				*otp++ = sign | e;
			}
			break;
		}
	}
}

void spack_int(char *dest, char *str, char len)
{
	unsigned _int64 i;
	char j;

	if (*str == '-')
		i = pow(2, 8*len) - atoi64(&str[1]);
	else
		i = atoi64(str);
	for (j = 0; j < len; ++j)
	{
		*dest++ = i & 255;
		i >>= 8;
	}
}

void pack_reg(char *str)
{
	*otp++ = *str - 'a';
}

void pack_regpair(char *str1, char *str2)
{
	*otp++ = ((*str1 - 'a') << 4) + (*str2 - 'a');
}

char label_exists(char *str)
{
	int j;

	for (j = 0; j < n_label; ++j)
		if (!strcmp(label[j].name, str))
			return label[j].filled;

	return -1;
}

int getlabel(char *str)
{
	int j;

	for (j = 0; j < n_label; ++j)
		if (!strcmp(label[j].name, str))
			return j;

	label = realloc(label, sizeof(labelt) * (n_label+1));
	strcpy(label[n_label].name, str);
	label[n_label].filled = 0;
	return n_label++;
}

void pack_label(char *str, char sz)
{
	labelr = realloc(labelr, sizeof(labelrt) * (n_labelr+1));
	labelr[n_labelr].id = getlabel(str);
	labelr[n_labelr].sz = sz;
	labelr[n_labelr++].loc = otp;
	otp += 4;
}

int do_file(char *path, char t)
{
	char ch, *temp, *instr, *tp, comment = 0, err[BS], *sp, synerr;
	int t_sz = BS;
	int j, i, ch1;
	unsigned char *cp;
	FILE *in;

	if (!(in = fopen(path, "r")))
		exiterr("Unable to open input file");

	temp = (char *)malloc(BS);
	instr = (char *)malloc(BS);
	tp = temp;

	while ((ch1 = fgetc(in)) != EOF)
	{
		ch = (char)ch1;

		if (tp - temp >= t_sz)
		{
			j = tp - temp;
			temp = (char *)realloc(temp, t_sz += BS);
			instr = (char *)realloc(instr, t_sz);
			tp = temp + j;
		}

		if (ch == '\n' || (!comment && (ch == '\\' || ch == ';')))
		{
			if (ch == '\n' && t)
				++line;
			if (!comment)
			{
				*tp = '\0';

				trim_spaces(temp);
				strcpy(instr, temp);
				if (*temp == '#')
				{
					_strlwr(temp);
					strcpy2(opcode, strtok(temp, " \t"));
					strcpy2(arg[0], strtok(NULL, " \t"));
					strcpy2(arg[1], strtok(NULL, ""));

					if (!strcmp(opcode, "#define"))
					{
						def = realloc(def, sizeof(deft) * (n_defs+1));
						strcpy2(def[n_defs].name, arg[0]);
						strcpy2(def[n_defs++].value, arg[1]);
					}
					else if (!strcmp(opcode, "#include"))
						do_file(arg[0], 0);
				}				
				else if (*temp == '.')
				{
					strcpy2(opcode, strtok(temp, " \t"));
					_strlwr(opcode);
					if (*opcode)
					{
						if (strcmp(opcode, ".str"))
						{
							i = 0;
							while (sp = strtok(NULL, ","))
							{
								if (i >= n_arg)
									arg = realloc(arg, ARGSZ * ++n_arg);
								strcpy(arg[i], sp);
								trim_spaces(arg[i]);
								_strlwr(arg[i++]);
							}
						}
						else
							strcpy2(arg[0], strtok(NULL, ""));
					}

					if (!strcmp(opcode, ".db"))
					{
						data = realloc(data, sizeof(datat) * (n_data+1));
						data[n_data].sz = i;
						data[n_data].dat = malloc(i);
						for (j = 0; j < i; ++j)
						{
							if (!is_int(arg[j]))
							{
								sprintf(err, "Invalid byte \"%s\"", arg[j]);
								print_err(err);
							}
							else
								spack_int(&(data[n_data].dat[j]), arg[j], 1);
						}
						n_data++;
					}
					else if (!strcmp(opcode, ".str"))
					{
						data[n_data].sz = strlen(arg[0]) + 1;
						strcpy(data[n_data++].dat, arg[0]);
					}
					else if (!strcmp(opcode, ".int"))
					{
						if (is_int(arg[0]))
						{
							ch1 = atoi(arg[0]);
							data[n_data].sz = (i - 1) * ch1;
							for (j = 1; j < i; ++j)
							{
								if (is_int(arg[j]))
									spack_int(&(data[n_data].dat[(j-1)*ch1]), arg[j], ch1);
								else
								{
									sprintf(err, "Invalid integer: \"%s\"", arg[j]);
									print_err(err);
								}
							}
							n_data++;
						}
						else
						{
							sprintf(err, "Invalid integer: \"%s\"", arg[0]);
							print_err(err);
						}
					}
				}
				else if (temp[strlen(temp)-1] == ':')
				{
					temp[strlen(temp)-1] = '\0';
					if (!is_label(temp))
					{
						sprintf(err, "Invalid label name: \"%s\"", temp);
						print_err(err);
					}
					else if (label_exists(temp) == 1)
					{
						sprintf(err, "Duplicate label: \"%s\"", temp);
						print_err(err);
					}
					else
					{
						i = getlabel(temp);
						label[i].loc = otp - out;
						label[i].filled = 1;
					}
				}					
				else if (*temp)
				{
					strcpy2(opcode, strtok(temp, " \t"));
					_strlwr(opcode);
					if (*opcode)
					{
						i = 0;
						while (sp = strtok(NULL, ","))
						{
							strcpy(arg[i], sp);
							trim_spaces(arg[i]);
							for (j = 0; j < n_defs; ++j)
							{
								if (!strcmp(def[j].name, arg[i]))
								{
									if (i >= n_arg)
										arg = realloc(arg, ARGSZ * ++n_arg);
									strcpy(arg[i], def[j].value);
									break;
								}
							}
							i++;
						}
					}

					synerr = 0;
					osz = 4;
					oos = SIZE_32;
					cp = &(opcode[strlen(opcode)-1]);
					if (*cp == '1' || *cp == '2' || *cp == '4' || *cp == '8')
					{
						switch (*cp)
						{
							case '1':
								oos = SIZE_8;
								break;
							case '2':
								oos = SIZE_16;
								break;
							case '4':
								oos = SIZE_32;
								break;
							case '8':
								oos = SIZE_64;
						}
						osz = *cp - '0';
						*cp = '\0';
					}

					if (!strcmp(opcode, "nop"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							*otp++ = OP_NOP;
					}
					else if (!strcmp(opcode, "ld"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]))
							{
								if (is_reg(arg[1]))
								{
									*otp++ = OP_LD_RR + oos;
									pack_regpair(arg[0], arg[1]);
								}
								else if (is_int(arg[1]))
								{
									*otp++ = OP_LD_RC + oos;
									pack_reg(arg[0]);
									pack_int(arg[1], osz);
								}
								else if (is_float(arg[1]))
								{
									*otp++ = OP_LD_RC + oos;
									pack_reg(arg[0]);
									cp = otp;
									pack_float(arg[1], oos);
								}
								else if (is_label(arg[1]))
								{
									*otp++ = OP_LD_RC + oos;
									pack_reg(arg[0]);
									pack_label(arg[1], osz);
								}
								else if (is_lptr(arg[1]))
								{
									squish(arg[1]);
									if (is_int(arg[1]))
									{
										*otp++ = OP_LD_RMC + oos;
										pack_reg(arg[0]);
										pack_int(arg[1], 4);
									}
									else if (is_label(arg[1]))
									{
										*otp++ = OP_LD_RMC + oos;
										pack_reg(arg[0]);
										pack_label(arg[1], 4);
									}
									else if (is_reg(arg[1]))
									{
										*otp++ = OP_LD_RMR + oos;
										pack_regpair(arg[0], arg[1]);
									}
									else
										synerr = 1;
								}
								else
									synerr = 1;
							}
							else if (is_lptr(arg[0]))
							{
								squish(arg[0]);
								if (is_int(arg[0]))
								{
									if (is_reg(arg[1]))
									{
										*otp++ = OP_LD_MCR + oos;
										pack_int(arg[0], 4);
										pack_reg(arg[1]);
									}
									else
										synerr = 1;
								}
								else if (is_reg(arg[0]))
								{
									if (is_reg(arg[1]))
									{
										*otp++ = OP_LD_MRR + oos;
										pack_regpair(arg[0], arg[1]);
									}
									else
										synerr = 1;
								}
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "add"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_ADD + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "fadd"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_FADD + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "sub"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_SUB + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "fsub"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_FSUB + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "mult"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_MULT + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "mults"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_MULTS + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "fmult"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_FMULT + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "div"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_DIV + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "divs"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_DIVS + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "fdiv"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_FDIV + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "mod"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_MOD + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "mods"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_MODS + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "sl"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_SL + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "sln"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_SLN + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "sr"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_SR + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "srn"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_SRN + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "neg"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_NEG + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "fneg"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_FNEG + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "or"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_OR + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "and"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_AND + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "xor"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_XOR + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "not"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_NOT + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "bor"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_BOR + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "band"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_BAND + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "bnot"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_BNOT + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "cp"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_CP + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "jp"))
					{
						if (i == 1)
						{
							if (is_int(arg[0]))
							{
								writeop(OP_JP);
								pack_int(arg[0], 4);
							}
							else if (is_label(arg[0]))
							{
								writeop(OP_JP);
								pack_label(arg[0], 4);
							}
							else if (is_reg(arg[0]))
							{
								writeop(OP_JP_R);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
						else if (!wrong_argnum(opcode, i, 2))
						{
							if (is_int(arg[1]))
							{
								if (!strcmp(arg[0], "z"))
								{
									writeop(OP_JP_Z);
									pack_int(arg[1], 4);
								}
								else if (!strcmp(arg[0], "nz"))
								{
									writeop(OP_JP_NZ);
									pack_int(arg[1], 4);
								}
								else if (!strcmp(arg[0], "c"))
								{
									writeop(OP_JP_C);
									pack_int(arg[1], 4);
								}
								else if (!strcmp(arg[0], "nc"))
								{
									writeop(OP_JP_NC);
									pack_int(arg[1], 4);
								}
								else
									synerr = 1;
							}
							else if (is_label(arg[1]))
							{
								if (!strcmp(arg[0], "z"))
								{
									writeop(OP_JP_Z);
									pack_label(arg[1], 4);
								}
								else if (!strcmp(arg[0], "nz"))
								{
									writeop(OP_JP_NZ);
									pack_label(arg[1], 4);
								}
								else if (!strcmp(arg[0], "c"))
								{
									writeop(OP_JP_C);
									pack_label(arg[1], 4);
								}
								else if (!strcmp(arg[0], "nc"))
								{
									writeop(OP_JP_NC);
									pack_label(arg[1], 4);
								}
								else
									synerr = 1;
							}
							else if (is_reg(arg[1]))
							{
								if (!strcmp(arg[0], "z"))
								{
									writeop(OP_JP_R_Z);
									pack_reg(arg[1]);
								}
								else if (!strcmp(arg[0], "nz"))
								{
									writeop(OP_JP_R_NZ);
									pack_reg(arg[1]);
								}
								else if (!strcmp(arg[0], "c"))
								{
									writeop(OP_JP_R_C);
									pack_reg(arg[1]);
								}
								else if (!strcmp(arg[0], "nc"))
								{
									writeop(OP_JP_R_NC);
									pack_reg(arg[1]);
								}
								else
									synerr = 1;
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "inc"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_INC + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "finc"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_FINC + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "dec"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_DEC + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "fdec"))
					{
						if (!wrong_argnum(opcode, i, 1)) {
							if (is_reg(arg[0])) {
								writeop(OP_FDEC + oos);
								pack_reg(arg[0]);
								}
							else {
								synerr = 1;
								}
							}
						}
					else if (!strcmp(opcode, "push")) {
						if (!wrong_argnum(opcode, i, 1)) {
							if (is_reg(arg[0])) {
								writeop(OP_PUSH + oos);
								pack_reg(arg[0]);
								}
							else {
								synerr = 1;
								}
							}
						}
					else if (!strcmp(opcode, "pop")) {
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_POP + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "call"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_CALL_R);
								pack_reg(arg[0]);
							}
							else if (is_int(arg[0]))
							{
								writeop(OP_CALL_C);
								pack_int(arg[0], 4);
							}
							else if (is_label(arg[0]))
							{
								writeop(OP_CALL_C);
								pack_label(arg[0], 4);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "ret"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							writeop(OP_RET);
					}
					else if (!strcmp(opcode, "retv"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_RETV + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "arg"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_int(arg[0]))
							{
								writeop(OP_ARG_C + oos);
								pack_int(arg[0], osz);
							}
							else if (is_float(arg[0]))
							{
								writeop(OP_ARG_C + oos);
								pack_float(arg[0], oos);
							}
							else if (is_reg(arg[0]))
							{
								writeop(OP_ARG_R + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "hcall"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_HCALL_R);
								pack_reg(arg[0]);
							}
							else if (is_int(arg[0]))
							{
								writeop(OP_HCALL_C);
								pack_int(arg[0], 2);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "hcpy"))
					{
						if (!wrong_argnum(opcode, i, 3))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]) && is_reg(arg[2]))
							{
								writeop(OP_HCPY);
								pack_regpair(arg[0], arg[1]);
								pack_reg(arg[2]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "hscpy"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_HSCPY);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "mcpy"))
					{
						if (!wrong_argnum(opcode, i, 3))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]) && is_reg(arg[2]))
							{
								writeop(OP_MCPY);
								pack_regpair(arg[0], arg[1]);
								pack_reg(arg[2]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "cpy"))
					{
						if (!wrong_argnum(opcode, i, 3))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]) && is_reg(arg[2]))
							{
								writeop(OP_CPY);
								pack_regpair(arg[0], arg[1]);
								pack_reg(arg[2]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "scpy"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_SCPY);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "dcpy"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_DCPY);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "pmode"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							writeop(OP_PMODE);
					}
					else if (!strcmp(opcode, "imode"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							writeop(OP_IMODE);
					}
					else if (!strcmp(opcode, "rmode"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							writeop(OP_RMODE);
					}
					else if (!strcmp(opcode, "omode"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							writeop(OP_OMODE);
					}
					else if (!strcmp(opcode, "emode"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							writeop(OP_EMODE);
					}
					else if (!strcmp(opcode, "sync"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							writeop(OP_SYNC);
					}
					else if (!strcmp(opcode, "rsend"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_RSEND);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "msend"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_MSEND);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "mbase"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_int(arg[0]))
							{
								writeop(OP_MBASE);
								pack_int(arg[0], 4);
							}
							else
							{
								synerr = 1;
							}
						}
					}
					else if (!strcmp(opcode, "malloc"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_MALLOC);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}						
					}
					else if (!strcmp(opcode, "free"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_FREE);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}						
					}
					else if (!strcmp(opcode, "lalloc"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_int(arg[0]))
							{
								writeop(OP_LALLOC);
								pack_int(arg[0], 4);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "lfree"))
					{
						if (wrong_argnum(opcode, i, 0))
							synerr = 1;
						else
							writeop(OP_LFREE);
					}
					else if (!strcmp(opcode, "laddr"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_int(arg[0]))
							{
								writeop(OP_LADDR);
								pack_int(arg[0], 4);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "cast"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_int(arg[0]))
							{
								writeop(OP_CAST);
								pack_int(arg[0], 4);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "test"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_TEST + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}						
					}
					else if (!strcmp(opcode, "stest"))
					{
						if (!wrong_argnum(opcode, i, 1))
						{
							if (is_reg(arg[0]))
							{
								writeop(OP_STEST + oos);
								pack_reg(arg[0]);
							}
							else
								synerr = 1;
						}						
					}
					else if (!strcmp(opcode, "fcp"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_FCP + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}						
					}
					else if (!strcmp(opcode, "scp"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_SCP + oos);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "gbf"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_int(arg[1]))
							{
								writeop(OP_GBF);
								pack_reg(arg[0]);
								pack_int(arg[1], 2);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "sbf"))
					{
						if (!wrong_argnum(opcode, i, 3))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]) &&
								is_int(arg[2]))
							{
								writeop(OP_SBF);
								pack_regpair(arg[0], arg[1]);
								pack_int(arg[2], 2);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "sgbf"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_int(arg[1]))
							{
								writeop(OP_SGBF);
								pack_reg(arg[0]);
								pack_int(arg[1], 2);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "ssbf"))
					{
						if (!wrong_argnum(opcode, i, 3))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]) &&
								is_int(arg[2]))
							{
								writeop(OP_SSBF);
								pack_regpair(arg[0], arg[1]);
								pack_int(arg[2], 2);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "zero"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_ZERO);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else if (!strcmp(opcode, "slen"))
					{
						if (!wrong_argnum(opcode, i, 2))
						{
							if (is_reg(arg[0]) && is_reg(arg[1]))
							{
								writeop(OP_SLEN);
								pack_regpair(arg[0], arg[1]);
							}
							else
								synerr = 1;
						}
					}
					else
					{
						sprintf(err, "Unknown opcode \"%s\"", opcode);
						print_err(err);
					}

					if (synerr)
					{
						sprintf(err, "Argument sytnax error: \"%s\"", instr);
						print_err(err);
					}

				}

					
			}
			comment = (ch == ';');
			tp = temp;
		}
		else if (!comment && ch != '\r')
			*tp++ = ch;
	}
	fclose(in);

	free(temp);
	free(instr);

	return 1;
}

int main(int argc, char *argv[])
{
	char temp[BS], *tp = temp, comment = 0, err[BS], *sp, ofname[BS] = "out.xc";
	int j, k, i;
	unsigned char *cp;

	if (argc == 1)
	{
		printf("\nxasm v0.1 xCode assembler\nCopyright Adam Chlipala (psion@tpu.org) 1998-99\n");
		puts("Usage: xasm <source.xas>");
		exit(0);
	}
	if (argc > 2)
		strcpy(ofname, argv[2]);

	data = malloc(0);
	label = malloc(0);
	labelr = malloc(0);
	def = malloc(0);
	arg = malloc(ARGSZ * START_NARG);
	do_file(argv[1], 1);

	if (errcount) return 1;

	// Label filling, AKA pass 2

	for (j = 0; j < n_label; ++j)
	{
		if (!label[j].filled)
		{
			sprintf(err, "Label not found: \"%s\"", label[j].name);
			print_err(err);
		}
	}

	if (errcount) return 1;

	for (j = 0; j < n_labelr; ++j)
	{
		i = label[labelr[j].id].loc;
		tp = labelr[j].loc;
		for (k = 0; k < labelr[j].sz; ++k)
		{
			*tp++ = i & 255;
			i >>= 8;
		}
	}

	sp = strtok(argv[1], ".");
	if (!(in = fopen(ofname, "wb")))
	{
		printf("Error writing %s; aborting\n", outfile);
		exit(1);
	}

	for (j = 0; j < n_data; ++j)
	{
		i = data[j].sz;
		for (k = 0; k < 4; ++k)
		{
			fputc(i & 255, in);
			i >>= 8;
		}
		i = data[j].sz;
		for (k = 0; k < i; ++k)
			fputc(data[j].dat[k], in);
	}

	fputc(0, in);
	fputc(0, in);
	fputc(0, in);
	fputc(0, in);
	for (cp = out; cp < otp; ++cp)
		fputc(*cp, in);
	fclose(in);

	for (j = 0; j < n_data; ++j)
		free(data[j].dat);
	free(data);
	free(label);
	free(labelr);
	free(def);
	free(arg);

	return 0;
}
